from datetime import datetime
import pytest

from psycopg2_mq import MQSource


def test_simple_call(model, dbsession):
    source = MQSource(dbsession=dbsession, model=model)
    with dbsession.begin():
        job_id = source.call('dummy', 'echo', {'message': 'hello world'})

    jobs = source.query_jobs.all()
    assert len(jobs) == 1
    job = jobs[0]
    assert job.queue == 'dummy'
    assert job.method == 'echo'
    assert job.args == {'message': 'hello world'}

    job2 = source.get_job(job_id)
    assert job is job2


def test_call_collapses_on_cursor(model, dbsession):
    source = MQSource(dbsession=dbsession, model=model)
    with dbsession.begin():
        job_id = source.call(
            'dummy',
            'echo',
            {'message': 'hello world original'},
            cursor_key='foo',
            collapse_on_cursor=True,
        )

    with dbsession.begin():
        job2_id = source.call(
            'dummy',
            'echo',
            {'message': 'hello world again'},
            cursor_key='foo',
            collapse_on_cursor=True,
        )

    assert job_id == job2_id

    job = source.get_job(job_id)
    assert job.args == {'message': 'hello world original'}


def test_retry_error(model, dbsession):
    source = MQSource(dbsession=dbsession, model=model)
    with dbsession.begin():
        job_id = source.call('dummy', 'echo', {'message': 'hello world'})

    with pytest.raises(RuntimeError) as ex:
        source.retry_job(job_id)

    assert 'job is not in a retryable state' in str(ex)


def test_retry_failed_job(model, dbsession):
    source = MQSource(dbsession=dbsession, model=model)
    with dbsession.begin():
        job_id = source.call('dummy', 'echo', {'message': 'hello world'})
        job = source.get_job(job_id)
        job.state = model.JobStates.FAILED
        job.start_time = datetime.utcnow()
        job.end_time = datetime.utcnow()

    job2_id = source.retry_job(job_id)
    assert job2_id != job_id

    job = source.get_job(job_id)
    assert job.state == model.JobStates.FAILED

    job = source.get_job(job2_id)
    assert job.state == model.JobStates.PENDING
    assert job.queue == 'dummy'
    assert job.method == 'echo'
    assert job.args == {'message': 'hello world'}



import argparse
import logging
import sqlalchemy as sa
import sqlalchemy.orm
import time

from psycopg2_mq import MQSource, make_default_model

metadata = sa.MetaData()
model = make_default_model(metadata)


def wait_for_job(source, job_id, *, max_wait=60):
    start = time.time()
    while True:
        if time.time() - start > max_wait:  # pragma: nocover
            raise Exception('timeout while waiting for job to finish', job_id)
        with source.dbsession.begin():
            job = source.find_job(job_id)
            if job.state in {
                source.model.JobStates.COMPLETED,
                source.model.JobStates.FAILED,
                source.model.JobStates.LOST,
            }:
                break
            time.sleep(0.1)


parser = argparse.ArgumentParser()
parser.add_argument('--url', required=True)
args = parser.parse_args()

logging.basicConfig(level=logging.DEBUG)

engine = sa.create_engine(args.url)
metadata.create_all(engine)

db = sa.orm.Session(engine)
source = MQSource(dbsession=db, model=model)

# continue with python -i



import pytest
from pytest_postgresql import factories
from pytest_postgresql.janitor import DatabaseJanitor
import sqlalchemy as sa
import sqlalchemy.orm

from psycopg2_mq import make_default_model

metadata = sa.MetaData()
_model = make_default_model(metadata)


def load_model(host, port, user, password, dbname):
    """
    Initialize the database schema with the default library model.
    """
    engine = sa.create_engine(f'postgresql://{user}:{password}@{host}:{port}/{dbname}')
    with engine.begin() as conn:
        metadata.create_all(bind=conn)


test_db_proc = factories.postgresql_proc(
    load=[load_model],
)


@pytest.fixture(scope='session')
def model():
    """
    The initialized database model objects.
    """
    return _model


@pytest.fixture
def dbengine(test_db_proc):
    """
    Create a new database per-test and return the SQLAlchemy engine.
    """
    user = test_db_proc.user
    password = test_db_proc.password
    host = test_db_proc.host
    port = test_db_proc.port
    dbname = test_db_proc.dbname
    template_dbname = test_db_proc.template_dbname

    # the janitor creates a new database for this specific test fixture,
    # copying data from the template database initialized in load_model above
    with DatabaseJanitor(
        version=test_db_proc.version,
        user=user,
        password=password,
        host=host,
        port=port,
        dbname=dbname,
        template_dbname=template_dbname,
    ):
        engine = sa.create_engine(
            f'postgresql://{user}:{password}@{host}:{port}/{dbname}'
        )
        try:
            yield engine
        finally:
            engine.dispose()


@pytest.fixture
def dbsession(dbengine):
    with sa.orm.Session(dbengine) as db:
        yield db



import argparse
import logging
import sqlalchemy as sa

from psycopg2_mq import MQWorker, make_default_model

metadata = sa.MetaData()
model = make_default_model(metadata)


class DummyQueue:
    def execute_job(self, job):
        print(job.id, job.queue, job.method, job.args)
        return {
            'queue': job.queue,
            'method': job.method,
            'args': job.args,
        }


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--url', required=True)
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG)

    engine = sa.create_engine(args.url)
    metadata.create_all(engine)

    queues = {
        'dummy': DummyQueue(),
    }

    worker = MQWorker(
        engine=engine,
        queues=queues,
        model=model,
    )
    worker.run()


if __name__ == '__main__':
    raise SystemExit(main())



from contextlib import contextmanager
import pytest
import threading
import time

from psycopg2_mq import MQSource, MQWorker


def test_simple_integration(model, dbsession, worker_proxy):
    worker_proxy.start(
        queues={
            'dummy': DummyQueue(),
        }
    )

    source = MQSource(dbsession=dbsession, model=model)
    with dbsession.begin():
        job_id = source.call('dummy', 'echo', {'message': 'hello world'})

    with wait_for_job(source, job_id) as job:
        assert job.state == model.JobStates.COMPLETED
        assert job.result == {
            'queue': 'dummy',
            'method': 'echo',
            'args': {'message': 'hello world'},
        }
        assert job.start_time is not None
        assert job.end_time is not None


def test_listener_integration(model, dbsession, worker_proxy):
    worker_proxy.start(
        queues={
            'dummy': DummyQueue(),
            'listener': DummyQueue(),
        },
        threads=2,
    )

    source = MQSource(dbsession=dbsession, model=model)
    with dbsession.begin():
        listener = source.add_listener(
            'mq_job_complete.dummy.echo', 'listener', 'listener_echo', {'a': 1}
        )
        listener_id = listener.id

        job_id = source.call('dummy', 'echo', {'message': 'hello world'})

    with wait_for_job(source, job_id) as job:
        assert job.state == model.JobStates.COMPLETED
        assert job.result == {
            'queue': 'dummy',
            'method': 'echo',
            'args': {'message': 'hello world'},
        }
        assert job.start_time is not None
        assert job.end_time is not None

        expected_event = {
            'name': 'mq_job_complete.dummy.echo',
            'listener_id': listener_id,
            'data': {
                'job_id': job.id,
                'result': job.result,
            },
        }

        job = (
            source.query_jobs.join(model.JobListenerLink)
            .filter(model.JobListenerLink.listener_id == listener_id)
            .one()
        )
        assert job.queue == 'listener'
        assert job.method == 'listener_echo'
        assert job.args == {'a': 1, 'event': expected_event}
        job_id = job.id

    with wait_for_job(source, job_id) as job:
        assert job.state == model.JobStates.COMPLETED
        assert job.result == {
            'queue': 'listener',
            'method': 'listener_echo',
            'args': {'a': 1, 'event': expected_event},
        }
        assert job.start_time is not None
        assert job.end_time is not None


class DummyQueue:
    def execute_job(self, job):
        return {
            'queue': job.queue,
            'method': job.method,
            'args': job.args,
        }


class WorkerProxy:
    worker = None
    thread = None

    def __init__(self, model, dbengine):
        self.model = model
        self.dbengine = dbengine

    def start(self, **kw):
        self.worker = MQWorker(
            engine=self.dbengine,
            model=self.model,
            capture_signals=False,
            timeout=10,
            **kw,
        )
        self.thread = threading.Thread(target=self.worker.run)
        self.thread.daemon = True
        self.thread.start()
        return self.worker

    def stop(self):
        self.worker.shutdown_gracefully()
        self.thread.join()


@pytest.fixture
def worker_proxy(model, dbengine):
    proxy = WorkerProxy(model, dbengine)
    try:
        yield proxy
    finally:
        proxy.stop()


@contextmanager
def wait_for_job(source, job_id, *, max_wait=60):
    start = time.time()
    while True:
        if time.time() - start > max_wait:  # pragma: nocover
            raise Exception('timeout while waiting for job to finish', job_id)
        with source.dbsession.begin():
            job = source.get_job(job_id)
            if job.state in {
                source.model.JobStates.COMPLETED,
                source.model.JobStates.FAILED,
                source.model.JobStates.LOST,
            }:
                yield job
                break
            time.sleep(0.1)



import argparse
from datetime import timedelta
import logging
import sqlalchemy as sa
import sqlalchemy.orm

from psycopg2_mq import MQSource, make_default_model

metadata = sa.MetaData()
model = make_default_model(metadata)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--url', required=True)
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG)

    engine = sa.create_engine(args.url)
    metadata.create_all(engine)

    with sa.orm.Session(engine) as db:
        source = MQSource(dbsession=db, model=model)

        listener = source.add_listener(
            'mq_job_complete.dummy.echo',
            'dummy',
            'listener_echo',
            {'source': 'echo-listener'},
            cursor_key='echo-listener',
            collapse_on_cursor=True,
            when=timedelta(seconds=30),
        )
        print('listener', listener)

        schedule = source.add_schedule(
            'dummy', 'echo', {'source': 'minutely'}, rrule='FREQ=MINUTELY'
        )
        print('schedule', schedule)

        schedule = source.add_schedule(
            'dummy', 'echo', {'source': 'secondly'}, rrule='FREQ=SECONDLY;INTERVAL=15'
        )
        print('schedule', schedule)
        db.commit()


if __name__ == '__main__':
    raise SystemExit(main())



import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.schema import CheckConstraint, Column, ForeignKey, Index
from sqlalchemy.types import BigInteger, Boolean, DateTime, Enum, Integer, Text


class Model:
    channel_prefix = 'mq_'

    def __init__(
        self,
        *,
        Job,
        JobStates,
        Lock,
        JobCursor,
        JobListener,
        JobSchedule,
        JobListenerLink=None,
        JobScheduleLink=None,
    ):
        self.Job = Job
        self.JobStates = JobStates
        self.JobCursor = JobCursor
        self.JobListener = JobListener
        self.JobSchedule = JobSchedule
        self.JobListenerLink = JobListenerLink
        self.JobScheduleLink = JobScheduleLink
        self.Lock = Lock


class JobStates:
    # constants used for the state parameter
    PENDING = 'pending'
    RUNNING = 'running'
    COMPLETED = 'completed'
    FAILED = 'failed'
    CANCELED = 'canceled'
    LOST = 'lost'


def make_default_model(metadata, JobStates=JobStates):
    Base = declarative_base(metadata=metadata)

    state_enum = Enum(
        JobStates.PENDING,
        JobStates.RUNNING,
        JobStates.COMPLETED,
        JobStates.FAILED,
        JobStates.CANCELED,
        JobStates.LOST,
        metadata=metadata,
        name='mq_job_state',
    )

    class Job(Base):
        __tablename__ = 'mq_job'

        id = Column(BigInteger, primary_key=True)
        start_time = Column(DateTime)
        end_time = Column(DateTime)
        state = Column(state_enum, nullable=False, index=True)

        created_time = Column(DateTime, nullable=False)
        scheduled_time = Column(DateTime, nullable=False, index=True)

        queue = Column(Text, nullable=False, index=True)
        method = Column(Text, nullable=False)
        args = Column(pg.JSONB, nullable=False)
        result = Column(pg.JSONB)

        cursor_key = Column(Text)
        cursor_snapshot = Column(pg.JSONB)
        collapsible = Column(Boolean)

        trace = Column(pg.JSONB)

        lock_id = Column(Integer)
        worker = Column(Text)

        listener_links = relationship(lambda: JobListenerLink, back_populates='job')
        listeners = relationship(
            lambda: JobListener,
            secondary=lambda: JobListenerLink.__table__,
            back_populates='jobs',
            viewonly=True,
        )

        schedule_links = relationship(lambda: JobScheduleLink, back_populates='job')
        schedules = relationship(
            lambda: JobSchedule,
            secondary=lambda: JobScheduleLink.__table__,
            back_populates='jobs',
            viewonly=True,
        )

        __table_args__ = (
            CheckConstraint(
                sa.or_(
                    state != JobStates.RUNNING,
                    lock_id != sa.null(),
                ),
                name='ck_mq_job_lock_id',
            ),
            Index(
                'uq_mq_job_pending_cursor_key_queue_method',
                cursor_key,
                queue,
                method,
                postgresql_where=sa.and_(
                    state == JobStates.PENDING,
                    collapsible == sa.true(),
                ),
                unique=True,
            ),
            Index(
                'uq_mq_job_running_cursor_key',
                cursor_key,
                postgresql_where=state == JobStates.RUNNING,
                unique=True,
            ),
        )

        def __repr__(self):
            return (
                '<Job('
                'id={0.id}'
                ', state="{0.state}", '
                ', scheduled_time={0.scheduled_time}'
                ', queue="{0.queue}"'
                ', method="{0.method}"'
                ', cursor_key="{0.cursor_key}"'
                ', collapsible="{0.collapsible}"'
                ')>'
            ).format(self)

    class JobCursor(Base):
        __tablename__ = 'mq_job_cursor'

        key = Column(Text, primary_key=True)
        properties = Column(pg.JSONB, default=dict, nullable=False)

        updated_job_id = Column(
            ForeignKey('mq_job.id', ondelete='set null', onupdate='cascade'),
        )
        updated_job = relationship(lambda: Job)

        def __repr__(self):
            return (
                f'<JobCursor(key="{self.key}", updated_job_id={self.updated_job_id})>'
            )

    class JobListener(Base):
        __tablename__ = 'mq_job_listener'

        id = Column(BigInteger, primary_key=True)

        created_time = Column(DateTime, nullable=False)

        is_enabled = Column(Boolean, nullable=False)

        event = Column(Text, nullable=False, index=True)

        queue = Column(Text, nullable=False, index=True)
        method = Column(Text, nullable=False)
        args = Column(pg.JSONB, nullable=False)

        # the event context will be added as an arg in the job via this key
        context_arg_key = Column(Text, nullable=False)

        # a duration in the future when the job should execute
        # relative to the event time, supports anything that the postgres
        # interval type can parse
        when = Column(pg.INTERVAL)

        cursor_key = Column(Text)
        collapse_on_cursor = Column(Boolean)

        job_links = relationship(lambda: JobListenerLink, back_populates='listener')
        jobs = relationship(
            lambda: Job,
            secondary=lambda: JobListenerLink.__table__,
            back_populates='listeners',
            viewonly=True,
        )

        def __repr__(self):
            return (
                '<JobListener('
                'id={0.id}'
                ', event="{0.event}"'
                ', queue="{0.queue}"'
                ', method="{0.method}"'
                ', when="{0.when}"'
                ', cursor_key="{0.cursor_key}"'
                ', collapse_on_cursor={0.collapse_on_cursor}'
                ', is_enabled={0.is_enabled}'
                '>'.format(self)
            )

    class JobSchedule(Base):
        __tablename__ = 'mq_job_schedule'

        id = Column(BigInteger, primary_key=True)

        created_time = Column(DateTime, nullable=False)

        is_enabled = Column(Boolean, nullable=False)

        rrule = Column(Text, nullable=False)

        queue = Column(Text, nullable=False, index=True)
        method = Column(Text, nullable=False)
        args = Column(pg.JSONB, nullable=False)

        cursor_key = Column(Text)
        # defaults to true if cursor_key is set
        collapse_on_cursor = Column(Boolean)

        next_execution_time = Column(DateTime, nullable=True)

        jobs = relationship(lambda: Job, back_populates='schedule')

        job_links = relationship(lambda: JobScheduleLink, back_populates='schedule')
        jobs = relationship(
            lambda: Job,
            secondary=lambda: JobScheduleLink.__table__,
            back_populates='schedules',
            viewonly=True,
        )

        def __repr__(self):
            return (
                '<JobSchedule('
                'id={0.id}'
                ', queue="{0.queue}"'
                ', method="{0.method}"'
                ', rrule="{0.rrule}"'
                ', cursor_key="{0.cursor_key}"'
                ', collapse_on_cursor={0.collapse_on_cursor}'
                ', is_enabled={0.is_enabled}'
                '>'.format(self)
            )

    class JobListenerLink(Base):
        __tablename__ = 'mq_job_listener_link'

        job_id = Column(
            ForeignKey('mq_job.id', ondelete='cascade', onupdate='cascade'),
            primary_key=True,
            nullable=False,
            index=True,
        )
        job = relationship(lambda: Job, back_populates='listener_links')

        listener_id = Column(
            ForeignKey('mq_job_listener.id', ondelete='cascade', onupdate='cascade'),
            primary_key=True,
            nullable=False,
            index=True,
        )
        listener = relationship(lambda: JobListener, back_populates='job_links')

        def __repr__(self):
            return (
                '<JobListenerLink('
                'job_id={0.job_id}'
                ', listener_id={0.listener_id}'
                '>'.format(self)
            )

    class JobScheduleLink(Base):
        __tablename__ = 'mq_job_schedule_link'

        job_id = Column(
            ForeignKey('mq_job.id', ondelete='cascade', onupdate='cascade'),
            primary_key=True,
            nullable=False,
            index=True,
        )
        job = relationship(lambda: Job, back_populates='schedule_links')

        schedule_id = Column(
            ForeignKey('mq_job_schedule.id', ondelete='cascade', onupdate='cascade'),
            primary_key=True,
            nullable=False,
            index=True,
        )
        schedule = relationship(lambda: JobSchedule, back_populates='job_links')

        def __repr__(self):
            return (
                '<JobScheduleLink('
                'job_id={0.job_id}'
                ', schedule_id={0.schedule_id}'
                '>'.format(self)
            )

    class Lock(Base):
        __tablename__ = 'mq_lock'

        queue = Column(Text, primary_key=True)
        key = Column(Text, primary_key=True)

        lock_id = Column(Integer, nullable=False)
        worker = Column(Text, nullable=False)

        def __repr__(self):
            return '<Lock(queue="{0.queue}", key="{0.key}">'.format(self)

    return Model(
        Job=Job,
        JobStates=JobStates,
        JobCursor=JobCursor,
        JobListener=JobListener,
        JobSchedule=JobSchedule,
        JobListenerLink=JobListenerLink,
        JobScheduleLink=JobScheduleLink,
        Lock=Lock,
    )



from dateutil.parser import parse as parse_datetime


def test_rrule():
    from psycopg2_mq.util import get_next_rrule_time

    now = parse_datetime('2023-04-01T06:00:00Z')
    target = get_next_rrule_time('FREQ=MONTHLY;BYDAY=1MO', now, after=now)
    assert target.isoformat() == '2023-04-03T06:00:00'



from copy import deepcopy
from datetime import datetime
import json
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import insert
from zope.sqlalchemy import mark_changed

from .util import datetime_to_int, get_next_rrule_time

log = __import__('logging').getLogger(__name__)


class MQSource:
    def __init__(
        self,
        *,
        dbsession,
        model,
        transaction_manager=None,
    ):
        self.dbsession = dbsession
        self.model = model
        self.transaction_manager = transaction_manager

    def call(
        self,
        queue,
        method,
        args,
        *,
        when=None,
        now=None,
        job_kwargs=None,
        cursor_key=None,
        collapse_on_cursor=False,
        conflict_resolver=None,
        schedule_ids=None,
        listener_ids=None,
        trace=None,
    ):
        """
        Dispatch a new job.

        :param collapse_on_cursor:
            If set to ``False``, then new jobs will always be created.
            When ``True`` and there is already a job on this cursor in the
            pending state with the same queue/method and marked collapsible then
            the ``conflict_resolver`` will be invoked with the existing job and
            the new request will be collapsed into the previous pending job.

            By default this is always ``False``.

        :param conflict_resolver:
            A callable that accepts the old job as its only argument. This will
            be invoked if ``collapse_on_cursor`` is ``True`` and a job already
            exists. The callable can modify the job object prior to returning
            to update any arguments prior to it running.

        """
        if now is None:
            now = datetime.utcnow()
        if when is None:
            when = now
        if args is None:
            args = {}
        if job_kwargs is None:
            job_kwargs = {}

        Job = self.model.Job
        JobListenerLink = self.model.JobListenerLink
        JobScheduleLink = self.model.JobScheduleLink
        JobStates = self.model.JobStates

        if collapse_on_cursor and not cursor_key:
            raise ValueError('cannot collapse a job that is not using a cursor')

        job_id = None
        notify = False
        # there is a race condition here in READ COMMITTED mode where
        # a job could start between the insert and the query, so if we
        # do not get a locked job record here we will loop until we either
        # insert a new row or get a locked one
        while True:
            job_id = self.dbsession.execute(
                insert(Job.__table__)
                .values(
                    {
                        Job.queue: queue,
                        Job.method: method,
                        Job.args: args,
                        Job.created_time: now,
                        Job.scheduled_time: when,
                        Job.state: JobStates.PENDING,
                        Job.cursor_key: cursor_key,
                        Job.collapsible: collapse_on_cursor,
                        Job.trace: trace,
                        **job_kwargs,
                    }
                )
                .on_conflict_do_nothing(
                    index_elements=[Job.cursor_key, Job.queue, Job.method],
                    index_where=sa.and_(
                        Job.state == JobStates.PENDING,
                        Job.collapsible == sa.true(),
                    ),
                )
                .returning(Job.id)
            ).scalar()
            if job_id is not None:
                if schedule_ids:
                    log.info(
                        'created new job=%s on queue=%s, method=%s from schedule=%s',
                        job_id,
                        queue,
                        method,
                        schedule_ids,
                    )
                elif listener_ids:
                    log.info(
                        'created new job=%s on queue=%s, method=%s from listener=%s',
                        job_id,
                        queue,
                        method,
                        listener_ids,
                    )
                else:
                    log.info(
                        'created new job=%s on queue=%s, method=%s',
                        job_id,
                        queue,
                        method,
                    )
                notify = True
                break

            # a job already exists, load it and resolve conflicts
            # XXX this will only occur on jobs with a cursor
            job = (
                self.dbsession.query(Job)
                .with_for_update()
                .filter(
                    Job.state == JobStates.PENDING,
                    Job.queue == queue,
                    Job.method == method,
                    Job.cursor_key == cursor_key,
                    Job.collapsible == sa.true(),
                )
                .first()
            )
            if job is not None:
                job_id = job.id
                log.info('joining existing job=%s', job_id)

                # the earlier scheduled_time should always win
                if when < job.scheduled_time:
                    job.scheduled_time = when
                    notify = True

                # invoke a callable to update properties on the job
                if conflict_resolver is not None:
                    conflict_resolver(job)

                    if self.transaction_manager:
                        mark_changed(self.dbsession)

                break

        if listener_ids and JobListenerLink is not None:
            self.dbsession.execute(
                insert(JobListenerLink.__table__)
                .values(
                    [
                        {
                            JobListenerLink.job_id: job_id,
                            JobListenerLink.listener_id: listener_id,
                        }
                        for listener_id in listener_ids
                    ]
                )
                .on_conflict_do_nothing(
                    index_elements=[
                        JobListenerLink.job_id,
                        JobListenerLink.listener_id,
                    ],
                )
            )

        if schedule_ids and JobScheduleLink is not None:
            self.dbsession.execute(
                insert(JobScheduleLink.__table__)
                .values(
                    [
                        {
                            JobScheduleLink.job_id: job_id,
                            JobScheduleLink.schedule_id: schedule_id,
                        }
                        for schedule_id in schedule_ids
                    ]
                )
                .on_conflict_do_nothing(
                    index_elements=[
                        JobScheduleLink.job_id,
                        JobScheduleLink.schedule_id,
                    ],
                )
            )

        if notify:
            epoch_seconds = datetime_to_int(when)
            payload = json.dumps({'j': job_id, 't': epoch_seconds})
            self.dbsession.execute(
                sa.select(
                    sa.func.pg_notify(f'{self.model.channel_prefix}{queue}', payload),
                )
            )
            # XXX notify is always true when the raw insert works above so we
            # handle the two scenarios (notify and raw insert) in which
            # mark_changed needs to be called with only a single call
            if self.transaction_manager:
                mark_changed(self.dbsession)

        return job_id

    @property
    def query_jobs(self):
        return self.dbsession.query(self.model.Job)

    def get_job(self, job_id, *, for_update=False, raises=True):
        if not for_update:
            job = self.dbsession.get(self.model.Job, job_id)
        else:
            job = self.query_jobs.filter_by(id=job_id).with_for_update().one_or_none()
        if not job and raises:
            raise LookupError('cannot find job', job_id)
        return job

    def retry_job(self, job_id, *, now=None):
        """
        Retry a job.

        This dispatches a new job, copying the metadata from a job that was
        failed or lost.

        """
        if now is None:
            now = datetime.utcnow()

        job = self.get_job(job_id)
        if job.state not in {
            self.model.JobStates.COMPLETED,
            self.model.JobStates.FAILED,
            self.model.JobStates.LOST,
        }:
            raise RuntimeError('job is not in a retryable state')

        if job.state == self.model.JobStates.LOST:
            self.fail_lost_job(job_id, now=now)

        kw = {}
        if self.model.JobListenerLink is not None:
            kw['listener_ids'] = [link.listener_id for link in job.listener_links]
        if self.model.JobScheduleLink is not None:
            kw['schedule_ids'] = [link.schedule_id for link in job.schedule_links]

        trace = deepcopy(job.trace or {})
        trace['mq_source_job_id'] = job_id

        new_job_id = self.call(
            job.queue,
            job.method,
            job.args,
            now=now,
            when=job.scheduled_time,
            cursor_key=job.cursor_key,
            collapse_on_cursor=False,
            trace=trace,
            **kw,
        )
        return new_job_id

    def fail_lost_job(self, job_id, *, now=None):
        """
        Mark a lost job as failed.

        """
        if now is None:
            now = datetime.utcnow()

        job = self.get_job(job_id, for_update=True)
        if job.state != self.model.JobStates.LOST:
            raise RuntimeError('job is not in a lost state')

        job.state = self.model.JobStates.FAILED
        if job.end_time is None:
            job.end_time = now

    def cancel_job(self, job_id, *, result=None, now=None):
        """
        Cancel a job.

        It must be ``PENDING``, ``FAILED``, or ``LOST``.

        """
        if now is None:
            now = datetime.utcnow()

        job = self.get_job(job_id, for_update=True)
        if job.state not in {
            self.model.JobStates.LOST,
            self.model.JobStates.PENDING,
            self.model.JobStates.FAILED,
        }:
            raise RuntimeError('job is not in a valid state')

        job.state = self.model.JobStates.CANCELED
        if result is not None:
            job.result = result
        if job.end_time is None:
            job.end_time = now
        log.info('marked job=%s canceled', job_id)

    def reload_scheduler(self, queue, *, now=None):
        """
        Trigger the scheduler to reload information about modified schedules.

        For example, this should be used if ``reload=False`` is used with
        :ref:`.add_schedule`, :ref:`.disable_schedule`, or
        :ref:`.enable_schedule` in order to batch changes prior to reloading.

        """
        if now is None:
            now = datetime.utcnow()
        epoch_seconds = datetime_to_int(now)
        payload = json.dumps({'s': None, 't': epoch_seconds})
        self.dbsession.execute(
            sa.select(
                sa.func.pg_notify(f'{self.model.channel_prefix}{queue}', payload),
            )
        )

    def add_schedule(
        self,
        queue,
        method,
        args,
        *,
        rrule,
        is_enabled=True,
        cursor_key=None,
        collapse_on_cursor=False,
        schedule_kwargs=None,
        now=None,
        reload=True,
    ):
        """
        Add a new schedule record to dispatch jobs at a specified frequency.

        """
        if now is None:
            now = datetime.utcnow()
        if args is None:
            args = {}
        if schedule_kwargs is None:
            schedule_kwargs = {}

        if collapse_on_cursor and not cursor_key:
            raise ValueError(
                'cannot collapse jobs from a schedule that is not using a cursor'
            )

        next_execution_time = get_next_rrule_time(rrule, now, now)

        schedule = self.model.JobSchedule(
            queue=queue,
            method=method,
            args=args,
            rrule=rrule,
            is_enabled=is_enabled,
            cursor_key=cursor_key,
            collapse_on_cursor=collapse_on_cursor,
            created_time=now,
            next_execution_time=next_execution_time,
            **schedule_kwargs,
        )
        self.dbsession.add(schedule)
        self.dbsession.flush()

        if reload:
            self.reload_scheduler(queue, now=schedule.next_execution_time)

        return schedule

    @property
    def query_schedules(self):
        return self.dbsession.query(self.model.JobSchedule)

    def get_schedule(self, schedule_id, *, for_update=False):
        q = self.query_schedules.filter_by(id=schedule_id)
        if for_update:
            q = q.with_for_update()
        return q.one()

    def disable_schedule(self, schedule_id):
        """
        Disable a schedule preventing any further jobs from being automatically
        dispatched.

        """
        schedule = self.get_schedule(schedule_id, for_update=True)
        if not schedule.is_enabled:
            log.info('schedule=%s is already disabled', schedule_id)
            return
        schedule.is_enabled = False
        schedule.next_execution_time = None
        log.debug('disabled schedule=%s', schedule_id)

    def enable_schedule(self, schedule_id, *, now=None, reload=True):
        """Enable a schedule for execution at its next scheduled time."""
        if now is None:
            now = datetime.utcnow()
        schedule = self.get_schedule(schedule_id, for_update=True)
        if schedule.is_enabled:
            log.info('schedule=%s is already enabled', schedule_id)
            return
        schedule.is_enabled = True
        schedule.next_execution_time = get_next_rrule_time(
            schedule.rrule, schedule.created_time, now
        )
        if reload and schedule.next_execution_time is not None:
            self.reload_scheduler(schedule.queue, now=schedule.next_execution_time)
        log.debug(
            'enabling schedule=%s, next execution time=%s',
            schedule_id,
            schedule.next_execution_time,
        )

    def call_schedule(self, schedule_id, *, now=None, reload=True, when=None):
        """
        Manually invoke a schedule, dispatching a job immediately.

        This is useful when a job may have been skipped because no schedulers
        were active at the time it was due to execute.

        """
        if now is None:
            now = datetime.utcnow()
        if when is None:
            when = now

        schedule = self.get_schedule(schedule_id, for_update=True)
        job_id = self.call(
            queue=schedule.queue,
            method=schedule.method,
            args=schedule.args,
            cursor_key=schedule.cursor_key,
            collapse_on_cursor=schedule.collapse_on_cursor,
            now=now,
            when=when,
            schedule_ids=[schedule.id],
        )

        if schedule.is_enabled:
            schedule.next_execution_time = get_next_rrule_time(
                schedule.rrule, schedule.created_time, now
            )

            if reload and schedule.next_execution_time is not None:
                self.reload_scheduler(schedule.queue, now=schedule.next_execution_time)

        return job_id

    def add_listener(
        self,
        event,
        queue,
        method,
        args,
        *,
        when=None,
        is_enabled=True,
        context_arg_key=None,
        cursor_key=None,
        collapse_on_cursor=False,
        listener_kwargs=None,
        now=None,
    ):
        """
        Add a new listener that can dispatch jobs when an event is emitted.

        """
        if now is None:
            now = datetime.utcnow()
        if args is None:
            args = {}
        if listener_kwargs is None:
            listener_kwargs = {}

        if collapse_on_cursor and not cursor_key:
            raise ValueError(
                'cannot collapse jobs from a schedule that is not using a cursor'
            )

        if not context_arg_key:
            context_arg_key = 'events' if collapse_on_cursor else 'event'

        listener = self.model.JobListener(
            event=event,
            queue=queue,
            method=method,
            args=args,
            context_arg_key=context_arg_key,
            when=when,
            is_enabled=is_enabled,
            cursor_key=cursor_key,
            collapse_on_cursor=collapse_on_cursor,
            created_time=now,
            **listener_kwargs,
        )
        self.dbsession.add(listener)
        self.dbsession.flush()

        return listener

    @property
    def query_listeners(self):
        return self.dbsession.query(self.model.JobListener)

    def get_listener(self, listener_id, *, for_update=False):
        q = self.query_listeners.filter_by(id=listener_id)
        if for_update:
            q = q.with_for_update()
        return q.one()

    def disable_listener(self, listener_id):
        """
        Disable a listener preventing any further jobs from being automatically
        dispatched.

        """
        listener = self.get_listener(listener_id, for_update=True)
        if not listener.is_enabled:
            log.info('listener=%s is already disabled', listener_id)
            return
        listener.is_enabled = False
        log.debug('disabled listener=%s', listener_id)

    def enable_listener(self, listener_id, *, now=None):
        """Enable a listener for execution when the next event is dispatched."""
        if now is None:
            now = datetime.utcnow()
        listener = self.get_listener(listener_id, for_update=True)
        if listener.is_enabled:
            log.info('listener=%s is already enabled', listener_id)
            return
        listener.is_enabled = True
        log.debug('enabling listener=%s', listener_id)

    def emit_event(self, name, data, *, now=None, trace=None):
        if now is None:
            now = datetime.utcnow()

        job_ids = []
        for listener in self.query_listeners.filter(
            self.model.JobListener.event == name,
            self.model.JobListener.is_enabled,
        ):
            when = now
            if listener.when is not None:
                when += listener.when

            arg_key = listener.context_arg_key

            event = {
                'name': name,
                'listener_id': listener.id,
                'data': data,
            }

            if listener.collapse_on_cursor:
                job_args = {
                    arg_key: [event],
                    **listener.args,
                }

                def conflict_resolver(job, arg_key=arg_key, event=event):
                    job.args = deepcopy(job.args)
                    job.args.setdefault(arg_key, []).append(event)

            else:
                conflict_resolver = None
                job_args = {
                    arg_key: event,
                    **listener.args,
                }

            job_id = self.call(
                queue=listener.queue,
                method=listener.method,
                args=job_args,
                cursor_key=listener.cursor_key,
                collapse_on_cursor=listener.collapse_on_cursor,
                conflict_resolver=conflict_resolver,
                now=now,
                when=when,
                listener_ids=[listener.id],
                trace=trace,
            )
            job_ids.append(job_id)
        return job_ids
